#include "IterationConfig.h"
#include "Config.h"
#include "Track.h" // for TrackBase::algoint_to_cstr, sigh

#include "nlohmann/json.hpp"

#include <fstream>
#include <regex>
#include <iostream>
#include <iomanip>

// Redefine to also support ordered_json ... we want to keep variable order in JSON save files.
#define ITCONF_DEFINE_TYPE_NON_INTRUSIVE(Type, ...)  \
    inline void to_json(nlohmann::json& nlohmann_json_j, const Type& nlohmann_json_t) { NLOHMANN_JSON_EXPAND(NLOHMANN_JSON_PASTE(NLOHMANN_JSON_TO, __VA_ARGS__)) } \
    inline void from_json(const nlohmann::json& nlohmann_json_j, Type& nlohmann_json_t) { NLOHMANN_JSON_EXPAND(NLOHMANN_JSON_PASTE(NLOHMANN_JSON_FROM, __VA_ARGS__)) } \
    inline void to_json(nlohmann::ordered_json& nlohmann_json_j, const Type& nlohmann_json_t) { NLOHMANN_JSON_EXPAND(NLOHMANN_JSON_PASTE(NLOHMANN_JSON_TO, __VA_ARGS__)) } \
    inline void from_json(const nlohmann::ordered_json& nlohmann_json_j, Type& nlohmann_json_t) { NLOHMANN_JSON_EXPAND(NLOHMANN_JSON_PASTE(NLOHMANN_JSON_FROM, __VA_ARGS__)) }

namespace mkfit {

// Begin AUTO code, some members commented out.

ITCONF_DEFINE_TYPE_NON_INTRUSIVE(mkfit::LayerControl,
  /* int   */   m_layer
)

ITCONF_DEFINE_TYPE_NON_INTRUSIVE(mkfit::SteeringParams,
  /* std::vector<LayerControl> */  m_layer_plan,
  /* int */ m_region,
  /* int */ m_fwd_search_pickup,
  /* int */ m_bkw_fit_last,
  /* int */ m_bkw_search_pickup

)

ITCONF_DEFINE_TYPE_NON_INTRUSIVE(mkfit::IterationLayerConfig,
  /* float */   m_select_min_dphi,
  /* float */   m_select_max_dphi,
  /* float */   m_select_min_dq,
  /* float */   m_select_max_dq,
  /* float */   c_dp_sf,
  /* float */   c_dp_0,
  /* float */   c_dp_1,
  /* float */   c_dp_2,
  /* float */   c_dq_sf,
  /* float */   c_dq_0,
  /* float */   c_dq_1,
  /* float */   c_dq_2,
  /* float */   c_c2_sf,
  /* float */   c_c2_0,
  /* float */   c_c2_1,
  /* float */   c_c2_2
)

ITCONF_DEFINE_TYPE_NON_INTRUSIVE(mkfit::IterationParams,
  /* int */   nlayers_per_seed,
  /* int */   maxCandsPerSeed,
  /* int */   maxHolesPerCand,
  /* int */   maxConsecHoles,
  /* float */   chi2Cut_min,
  /* float */   chi2CutOverlap,
  /* float */   pTCutOverlap,
  /* float */   c_ptthr_hpt,
  /* float */   c_drmax_bh,
  /* float */   c_dzmax_bh,
  /* float */   c_drmax_eh,
  /* float */   c_dzmax_eh,
  /* float */   c_drmax_bl,
  /* float */   c_dzmax_bl,
  /* float */   c_drmax_el,
  /* float */   c_dzmax_el,
  /* int */   minHitsQF,
  /* float */   fracSharedHits,
  /* float */   drth_central,
  /* float */   drth_obarrel,
  /* float */   drth_forward

)

ITCONF_DEFINE_TYPE_NON_INTRUSIVE(mkfit::IterationConfig,
  /* int */                      m_iteration_index,
  /* int */                      m_track_algorithm,
  /* bool */                     m_requires_seed_hit_sorting,
  /* bool */                     m_requires_quality_filter,
  /* bool */                     m_requires_dupclean_tight,
  /* bool */                     m_backward_search,
  /* bool */                     m_backward_drop_seed_hits,
  /* int */                      m_backward_fit_min_hits,
  /* mkfit::IterationParams */   m_params,
  /* mkfit::IterationParams */   m_backward_params,
  /* int */                      m_n_regions,
  /* vector<int> */              m_region_order,
  /* vector<mkfit::SteeringParams> */         m_steering_params,
  /* vector<mkfit::IterationLayerConfig> */   m_layer_configs
  // /* function<void(const TrackerInfo&,const TrackVec&,const EventOfHits&,IterationSeedPartition&)> */   m_partition_seeds
)

ITCONF_DEFINE_TYPE_NON_INTRUSIVE(mkfit::IterationsInfo,
  /* vector<mkfit::IterationConfig> */   m_iterations
)

// End AUTO code.


// ============================================================================
// ConfigJsonPatcher
// ============================================================================

ConfigJsonPatcher::ConfigJsonPatcher(bool verbose) : m_verbose(verbose) {}

ConfigJsonPatcher::~ConfigJsonPatcher() { release_json(); }

void ConfigJsonPatcher::release_json()
{
    if (m_owner) delete m_json;
    m_json  = nullptr;
    m_owner = false;
}

std::string ConfigJsonPatcher::get_abs_path() const
{
    std::string s;
    s.reserve(64);
    for (auto &p : m_path_stack) s += p;
    return s;
}

std::string ConfigJsonPatcher::exc_hdr(const char* func) const
{
    std::string s;
    s.reserve(128);
    s = "ConfigJsonPatcher";
    if (func) { s += "::"; s += func; }
    s += " '";
    s += get_abs_path();
    s += "' ";
    return s;
}

template<class T> void ConfigJsonPatcher::Load(const T& o)
{
    release_json();
    m_json  = new nlohmann::json;
    *m_json = o;
    m_owner = true;
    cd_top();
}
template void ConfigJsonPatcher::Load<IterationsInfo> (const IterationsInfo  &o);
template void ConfigJsonPatcher::Load<IterationConfig>(const IterationConfig &o);

template<class T> void ConfigJsonPatcher::Save(T& o)
{
    from_json(*m_json, o);
}
template void ConfigJsonPatcher::Save<IterationConfig>(IterationConfig &o);

// Must not bork the IterationConfig elements of IterationsInfo ... default
// deserializator apparently reinitializes the vectors with defaults c-tors.
template <> void ConfigJsonPatcher::Save<IterationsInfo> (IterationsInfo &o)
{
    auto &itc_arr = m_json->at("m_iterations");
    for (int i = 0; i < o.size(); ++i)
    {
        from_json(itc_arr[i], o[i]);
    }
}

void ConfigJsonPatcher::cd(const std::string& path)
{
    nlohmann::json::json_pointer jp(path);
    m_json_stack.push_back(m_current);
    m_path_stack.push_back(path);
    m_current = & m_current->at(jp);
}

void ConfigJsonPatcher::cd_up(const std::string& path)
{
    if (m_json_stack.empty()) throw std::runtime_error("JSON stack empty on cd_up");

    m_current = m_json_stack.back();
    m_json_stack.pop_back();
    m_path_stack.pop_back();
    if ( ! path.empty()) cd(path);
}

void ConfigJsonPatcher::cd_top(const std::string& path)
{
    m_current = m_json;
    m_json_stack.clear();
    m_path_stack.clear();
    if ( ! path.empty()) cd(path);
}

template<typename T>
void ConfigJsonPatcher::replace(const std::string& path, T val)
{
    nlohmann::json::json_pointer jp(path);
    m_current->at(jp) = val;
}
template void ConfigJsonPatcher::replace<int>   (const std::string& path, int    val);
template void ConfigJsonPatcher::replace<float> (const std::string& path, float  val);
template void ConfigJsonPatcher::replace<double>(const std::string& path, double val);

template<typename T>
void ConfigJsonPatcher::replace(int first, int last, const std::string& path, T val)
{
    nlohmann::json::json_pointer jp(path);
    for (int i = first; i <= last; ++i)
    {
        m_current->at(i).at(jp) = val;
    }
}
template void ConfigJsonPatcher::replace<int>   (int first, int last, const std::string& path, int    val);
template void ConfigJsonPatcher::replace<float> (int first, int last, const std::string& path, float  val);
template void ConfigJsonPatcher::replace<double>(int first, int last, const std::string& path, double val);

nlohmann::json& ConfigJsonPatcher::get(const std::string& path)
{
    nlohmann::json::json_pointer jp(path);
    return m_current->at(jp);
}

int ConfigJsonPatcher::replace(const nlohmann::json &j)
{
    if (j.is_null()) throw std::runtime_error(exc_hdr(__func__) + "null not expected");

    if (j.is_boolean() || j.is_number() || j.is_string())
    {
        throw std::runtime_error(exc_hdr(__func__) + "value not expected on this parsing level");
    }

    int n_replaced = 0;

    if (j.is_object())
    {
        static const std::regex index_range_re("^\\[(\\d+)..(\\d+)\\]$", std::regex::optimize);

        for (auto& [key, value] : j.items())
        {
            std::smatch m;
            std::regex_search(key, m, index_range_re);

            if (m.size() == 3)
            {
                if ( ! m_current->is_array())  throw std::runtime_error(exc_hdr(__func__) + "array range encountered when current json is not an array");
                int first = std::stoi(m.str(1));
                int last  = std::stoi(m.str(2));
                for (int i = first; i <= last; ++i)
                {
                    std::string s("/");
                    s += std::to_string(i);
                    cd(s);
                    if (value.is_array())
                    {
                        for (auto &el : value)
                            n_replaced += replace(el);
                    }
                    else
                    {
                        n_replaced += replace(value);
                    }
                    cd_up();
                }
            }
            else if (value.is_array() || value.is_object())
            {
                std::string s("/");
                s += key;
                cd(s);
                n_replaced += replace(value);
                cd_up();
            }
            else if (value.is_number() || value.is_boolean() || value.is_string())
            {
                std::string s("/");
                s += key;
                nlohmann::json::json_pointer jp(s);
                if (m_current->at(jp) != value)
                {
                    if (m_verbose)
                        std::cout << "  " << get_abs_path() << s << ": " << m_current->at(jp) << " -> " << value << "\n";

                    m_current->at(jp) = value;
                    ++n_replaced;
                }
                // else {
                //     if (m_verbose)
                //         std::cout << "  " << get_abs_path() << s << ": " << m_current->at(jp) << " -> unchanged\n";
                // }
            }
            else
            {
                throw std::runtime_error(exc_hdr(__func__) + "unexpected value type");
            }
        }
    }
    else if (j.is_array() && j.empty())
    {}
    else if (j.is_array())
    {
        // Arrays are somewhat tricky.
        // At the moment all elements are expected to be objects.
        //    This means arrays of basic types are not supported (like layer index arrays).
        //    Should not be too hard to add support for this.
        // Now, the objects in the array can be of two kinds:
        // a) Their keys can be json_pointer strings starting with numbers or ranges [i_low..i_high].
        // b) They can be actual elements of the array. In this case we require the length of
        //    the array to be equal to existing length in the configuration.
        // It is not allowed for these two kinds to mix.

        // Determine the kind of array: json_ptr or object

        static const std::regex index_re("^(?:\\[\\d+..\\d+\\]|\\d+(?:/.*)?)$", std::regex::optimize);

        bool has_index = false, has_plain = false;
        for (int i = 0; i < (int) j.size(); ++i)
        {
            const nlohmann::json &el = j[i];

            if ( ! el.is_object()) throw std::runtime_error(exc_hdr(__func__) + "array elements expected to be objects");

            for (nlohmann::json::const_iterator it = el.begin(); it != el.end(); ++it)
            {
                if (std::regex_search(it.key(), index_re)) {
                    has_index = true;
                    if (has_plain) throw std::runtime_error(exc_hdr(__func__) + "indexed array entry following plain one");
                } else {
                    has_plain = true;
                    if (has_index) throw std::runtime_error(exc_hdr(__func__) + "plain array entry following indexed one");
                }
            }
        }
        if (has_index)
        {
            for (auto& element : j)
            {
                n_replaced += replace(element);
            }
        }
        else
        {
            if ( m_current && ! m_current->is_array())
                throw std::runtime_error(exc_hdr(__func__) + "plain array detected when current is not an array");
            if (m_current->size() != j.size())
                throw std::runtime_error(exc_hdr(__func__) + "plain array of different size than at current pos");

            std::string s;
            for (int i = 0; i < (int) j.size(); ++i)
            {
                s  = "/";
                s += std::to_string(i);
                cd(s);
                n_replaced += replace(j[i]);
                cd_up();
            }
        }
    }
    else
    {
        throw std::runtime_error(exc_hdr(__func__) + "unexpected json type");
    }

    return n_replaced;
}

std::string ConfigJsonPatcher::dump(int indent)
{
    return m_json->dump(indent);
}


// ============================================================================
// ConfigJson_Patch_File steering function
// ============================================================================
/*
    See example JSON patcher input: "mkFit/config-parse/test.json"

    The file can contain several valid JSON dumps in sequence.

    '/' character can be used to descend more than one level at a time.

    A number can be used to specify an array index. This can be combined with
    the '/' syntax.

    "[first,last]" key (as string) can be used to denote a range of array
    elements. Such a key must not be combined with a '/' syntax.
*/

namespace
{
    // Open file for writing, throw exception on failure.
    void open_ofstream(std::ofstream &ofs, const std::string &fname, const char *pfx=0)
    {
        ofs.open(fname, std::ofstream::trunc);
        if ( ! ofs)
        {
            char m[2048];
            snprintf(m, 2048, "%s%sError opening %s for write: %m", pfx ? pfx : "", pfx ? " " : "", fname.c_str());
            throw std::runtime_error(m);
        }
    }

    // Open file for reading, throw exception on failure.
    void open_ifstream(std::ifstream &ifs, const std::string &fname, const char *pfx=0)
    {
        ifs.open(fname);
        if ( ! ifs)
        {
            char m[2048];
            snprintf(m, 2048, "%s%sError opening %s for read: %m", pfx ? pfx : "", pfx ? " " : "", fname.c_str());
            throw std::runtime_error(m);
        }
    }

    // Skip white-space, return true if more characters are available, false if eof.
    bool skipws_ifstream(std::ifstream &ifs)
    {
        while (std::isspace(ifs.peek())) ifs.get();
        return ! ifs.eof();
    }
}

void ConfigJson_Patch_Files(IterationsInfo &its_info, const std::vector<std::string> &fnames,
                           ConfigJsonPatcher::PatchReport *report)
{
    ConfigJsonPatcher cjp(Config::json_verbose);
    cjp.Load(its_info);

    ConfigJsonPatcher::PatchReport rep;

    for (auto &fname : fnames)
    {
        std::ifstream ifs;
        open_ifstream(ifs, fname, __func__);

        if (Config::json_verbose)
        {
            printf("%s begin reading from file %s.\n", __func__, fname.c_str());
        }

        int n_read = 0, n_tot_replaced = 0;
        while (skipws_ifstream(ifs))
        {
            nlohmann::json j;
            ifs >> j;
            ++n_read;

            if (Config::json_verbose)
            {
                std::cout << " Read JSON entity " << n_read << " -- applying patch:\n";
                // std::cout << j.dump(3) << "\n";
            }

            int n_replaced = cjp.replace(j);

            if (Config::json_verbose)
            {
                std::cout << " Replaced " << n_replaced << " entries.\n";
            }
            cjp.cd_top();
            n_tot_replaced += n_replaced;
        }

        if (Config::json_verbose)
        {
            printf("%s read %d JSON entities from file %s, replaced %d parameters.\n",
                __func__, n_read, fname.c_str(), n_tot_replaced);
        }

        ifs.close();

        rep.inc_counts(1, n_read, n_tot_replaced);
    }

    if (rep.n_replacements > 0)
    {
        cjp.Save(its_info);
    }

    if (report) report->inc_counts(rep);
}

std::unique_ptr<IterationConfig>
ConfigJson_PatchLoad_File(const IterationsInfo &its_info, const std::string &fname,
                          ConfigJsonPatcher::PatchReport *report)
{
    ConfigJsonPatcher::PatchReport rep;

    std::ifstream ifs;
    open_ifstream(ifs, fname, __func__);

    if (Config::json_verbose)
    {
        printf("%s begin reading from file %s.\n", __func__, fname.c_str());
    }

    if ( ! skipws_ifstream(ifs)) throw std::runtime_error("empty file");

    nlohmann::json j;
    ifs >> j;
    int track_algo = j["m_track_algorithm"];

    int iii = -1;
    for (int i = 0; i < its_info.size(); ++i)
    {
        if (its_info[i].m_track_algorithm == track_algo) {
            iii = i;
            break;
        }
    }
    if (iii == -1) throw std::runtime_error("matching IterationConfig not found");

    if (Config::json_verbose)
    {
        std::cout << " Read JSON entity, Iteration index is " << iii
                  << " -- cloning and applying JSON patch:\n";
    }

    IterationConfig *icp = new IterationConfig( its_info[iii] );
    IterationConfig &ic  = * icp;

    ConfigJsonPatcher cjp(Config::json_verbose);
    cjp.Load(ic);

    int n_replaced = cjp.replace(j);

    cjp.cd_top();

    if (Config::json_verbose)
    {
        printf("%s read 1 JSON entity from file %s, replaced %d parameters.\n",
            __func__, fname.c_str(), n_replaced);
    }

    ifs.close();

    rep.inc_counts(1, 1, n_replaced);

    if (rep.n_replacements > 0)
    {
        cjp.Save(ic);
    }

    if (report) report->inc_counts(rep);

    return std::unique_ptr<IterationConfig>( icp );
}

std::unique_ptr<IterationConfig>
ConfigJson_Load_File(const std::string &fname)
{
    std::ifstream ifs;
    open_ifstream(ifs, fname, __func__);

    if (Config::json_verbose)
    {
        printf("%s begin reading from file %s.\n", __func__, fname.c_str());
    }

    if ( ! skipws_ifstream(ifs)) throw std::runtime_error("empty file");

    nlohmann::json j;
    ifs >> j;

    if (Config::json_verbose)
    {
        std::cout << " Read JSON entity, iteration index is " << j["m_iteration_index"]
                  << ", track algorithm is " << j["m_track_algorithm"]
                  << ". Instantiating IterationConfig object and over-laying it with JSON.\n";
    }

    IterationConfig *icp = new IterationConfig();

    from_json(j, *icp);

    return std::unique_ptr<IterationConfig>( icp );
}


// ============================================================================
// Save each IterationConfig into a separate json file
// ============================================================================

void ConfigJson_Save_Iterations(IterationsInfo &its_info, const std::string &fname_fmt,
                                bool include_iter_info_preamble)
{
    bool has_pct_d = fname_fmt.find("%d") != std::string::npos;
    bool has_pct_s = fname_fmt.find("%s") != std::string::npos;

    assert( (has_pct_d || has_pct_s) && "JSON save filename-format must include a %d or %s substring");
    assert(!(has_pct_d && has_pct_s) && "JSON save filename-format must include only one of %d or %s substrings");

    for (int ii = 0; ii < its_info.size(); ++ii)
    {
        const IterationConfig &itconf = its_info[ii];

        char fname[1024];
        if (has_pct_d)
            snprintf(fname, 1024, fname_fmt.c_str(), ii);
        else
            snprintf(fname, 1024, fname_fmt.c_str(), TrackBase::algoint_to_cstr(itconf.m_track_algorithm));

        std::ofstream ofs;
        open_ofstream(ofs, fname, __func__);

        if (include_iter_info_preamble)
        {
            ofs << "{ \"m_iterations/" << ii << "\": ";
        }

        nlohmann::ordered_json j;
        to_json(j, itconf);

        ofs << std::setw(1);
        ofs << j;

        if (include_iter_info_preamble)
        {
            ofs << " }";
        }

        ofs << "\n";
        ofs.close();
    }
}

void ConfigJson_Dump(IterationsInfo &its_info)
{
    nlohmann::ordered_json j = its_info;
    std::cout << j.dump(3) << "\n";
}

// ============================================================================
// Tests for ConfigJson stuff
// ============================================================================

void ConfigJson_Test_Direct(IterationConfig &it_cfg)
{
    using nlohmann::json;

    std::string lojz("/m_select_max_dphi");

    json j = it_cfg;
    std::cout << j.dump(1) << "\n";

    std::cout << "Layer 43, m_select_max_dphi = " << j["/m_layer_configs/43/m_select_max_dphi"_json_pointer] << "\n";
    std::cout << "Patching it to pi ...\n";
    json p = R"([
        { "op": "replace", "path": "/m_layer_configs/43/m_select_max_dphi", "value": 3.141 }
    ])"_json;
    j = j.patch(p);
    std::cout << "Layer 43, m_select_max_dphi = " << j["/m_layer_configs/43/m_select_max_dphi"_json_pointer] << "\n";

    auto &jx = j["/m_layer_configs/60"_json_pointer];
    // jx["m_select_max_dphi"] = 99.876;
    json::json_pointer jp(lojz);
    jx[jp] = 99.876;

    // try loading it back, see what happens to vector m_layer_configs.

    from_json(j, it_cfg);
    printf("Layer 43 : m_select_max_dphi = %f, size_of_layer_vec=%d, m_n_regions=%d, size_of_steering_params=%d\n",
           it_cfg.m_layer_configs[43].m_select_max_dphi, (int) it_cfg.m_layer_configs.size(),
           it_cfg.m_n_regions, (int) it_cfg.m_steering_params.size());

    printf("Layer 60 : m_select_max_dphi = %f, size_of_layer_vec=%d, m_n_regions=%d, size_of_steering_params=%d\n",
           it_cfg.m_layer_configs[60].m_select_max_dphi, (int) it_cfg.m_layer_configs.size(),
           it_cfg.m_n_regions, (int) it_cfg.m_steering_params.size());

    // try accessing something that does not exist

    // std::cout << "Non-existent path " << j["/m_layer_configs/143/m_select_max_dphi"_json_pointer] << "\n";

    auto &x = j["/m_layer_configs"_json_pointer];
    std::cout << "Typename /m_layer_configs " << x.type_name() << "\n";
    auto &y = j["/m_layer_configs/143"_json_pointer];
    std::cout << "Typename /m_layer_configs/143 " << y.type_name() << ", is_null=" << y.is_null() << "\n";
}

void ConfigJson_Test_Patcher(IterationConfig &it_cfg)
{
    ConfigJsonPatcher cjp;
    cjp.Load(it_cfg);

    std::cout << cjp.dump(1) << "\n";

    {
        cjp.cd("/m_layer_configs/43/m_select_max_dphi");
        std::cout << "Layer 43, m_select_max_dphi = " << cjp.get("") << "\n";
        std::cout << "Setting it to pi ...\n";
        cjp.replace("", 3.141);
        cjp.cd_top();
        std::cout << "Layer 43, m_select_max_dphi = " << cjp.get("/m_layer_configs/43/m_select_max_dphi") << "\n";
    }
    {
        std::cout << "Replacing layer 60 m_select_max_dphi with full path\n";
        cjp.replace("/m_layer_configs/60/m_select_max_dphi", 99.876);
    }
    try
    {
        std::cout << "Trying to replace an non-existent array entry\n";
        cjp.replace("/m_layer_configs/1460/m_select_max_dphi", 666.666);
    }
    catch (std::exception &exc)
    {
        std::cout << "Caugth exception: " << exc.what() << "\n";
    }
    try
    {
        std::cout << "Trying to replace an non-existent object entry\n";
        cjp.replace("/m_layer_configs/1/moo_select_max_dphi", 666.666);
    }
    catch (std::exception &exc)
    {
        std::cout << "Caugth exception: " << exc.what() << "\n";
    }
    {
        std::cout << "Replacing m_select_max_dphi on layers 1 to 3 to 7.7\n";
        cjp.cd("/m_layer_configs");
        cjp.replace(1, 3, "/m_select_max_dphi", 7.7);
        cjp.cd_top();
    }

    // try getting it back into c++, see what happens to vector m_layer_configs.

    cjp.Save(it_cfg);

    printf("Layer 43: m_select_max_dphi = %f, size_of_layer_vec=%d, m_n_regions=%d, size_of_steering_params=%d\n",
           it_cfg.m_layer_configs[43].m_select_max_dphi, (int) it_cfg.m_layer_configs.size(),
           it_cfg.m_n_regions, (int) it_cfg.m_steering_params.size());

    printf("Layer 60: m_select_max_dphi = %f\n", it_cfg.m_layer_configs[60].m_select_max_dphi);
    for (int i = 0; i < 5; ++i)
        printf("Layer %2d: m_select_max_dphi = %f\n", i, it_cfg.m_layer_configs[i].m_select_max_dphi);

    // try accessing something that does not exist

    // std::cout << "Non-existent path " << j["/m_layer_configs/143/m_select_max_dphi"_json_pointer] << "\n";

    auto &j = cjp.get("");

    auto &x = j["/m_layer_configs"_json_pointer];
    std::cout << "Typename /m_layer_configs " << x.type_name() << "\n";
    auto &y = j["/m_layer_configs/143"_json_pointer];
    std::cout << "Typename /m_layer_configs/143 " << y.type_name() << ", is_null=" << y.is_null() << "\n";
}

}
